/*-
 * Copyright (c) 1990, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Olson.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)rec_open.c	8.4 (Berkeley) 9/7/93";
#endif /* LIBC_SCCS and not lint */

#include <sys/types.h>

#ifndef WIN32
#include <sys/mman.h>
#endif

#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <unistd.h>

#define	__DBINTERFACE_PRIVATE
#include <db.h>
#include "recno.h"

DB *
__rec_open(fname, flags, mode, openinfo, dflags)
	const char *fname;
	int flags, mode, dflags;
	const RECNOINFO *openinfo;
{
	BTREE *t;
	BTREEINFO btopeninfo;
	DB *dbp;
	PAGE *h;
	struct stat sb;
	int sverrno;

#ifndef WIN32
        int rfd;
#else
        HANDLE Hrfd,Hfilemap;
        int ntaflags,ntoflags;
        BY_HANDLE_FILE_INFORMATION BHFI;

#endif

#ifndef WIN32
	/* Open the user's file -- if this fails, we're done. */
	if (fname != NULL && (rfd = open(fname, flags, mode)) < 0)
		return (NULL);
#else
    /* set Windows NT flags appropriately */

        ntoflags = OPEN_ALWAYS;
        ntaflags = 0;
        ntaflags |= (flags & O_WRONLY) ? GENERIC_WRITE : 0;
        ntaflags |= (flags & O_RDONLY) ? GENERIC_READ : 0;
        ntaflags |= (flags & O_RDWR) ? (GENERIC_WRITE|GENERIC_READ) : 0;
        ntoflags |= (flags & O_TRUNC) ? CREATE_NEW : 0;

        printf("rec_open: about to open file %s\n",fname);

        if (fname != NULL && (Hrfd = 
        CreateFile(fname, ntaflags,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
          NULL, ntoflags, FILE_ATTRIBUTE_NORMAL,
            NULL)) == INVALID_HANDLE_VALUE)
        return ((DB *)NULL);
#endif

	/* Create a btree in memory (backed by disk). */
	dbp = NULL;
	if (openinfo) {
		if (openinfo->flags & ~(R_FIXEDLEN | R_NOKEY | R_SNAPSHOT))
			goto einval;
		btopeninfo.flags = 0;
		btopeninfo.cachesize = openinfo->cachesize;
		btopeninfo.maxkeypage = 0;
		btopeninfo.minkeypage = 0;
		btopeninfo.psize = openinfo->psize;
		btopeninfo.compare = NULL;
		btopeninfo.prefix = NULL;
		btopeninfo.lorder = openinfo->lorder;
#ifndef WIN32
		dbp = __bt_open(openinfo->bfname,
		    O_RDWR, S_IRUSR | S_IWUSR, &btopeninfo, dflags);
#else
		dbp = __bt_open(openinfo->bfname,
                        O_RDWR, S_IRUSR | S_IWUSR,
                    &btopeninfo, dflags);
#endif
	} else

		dbp = __bt_open(NULL, 
                    O_RDWR, S_IRUSR | S_IWUSR, 
                    NULL, dflags);

	if (dbp == NULL)
		goto err;

	/*
	 * Some fields in the tree structure are recno specific.  Fill them
	 * in and make the btree structure look like a recno structure.  We
	 * don't change the bt_ovflsize value, it's close enough and slightly
	 * bigger.
	 */
	t = dbp->internal;
	if (openinfo) {
		if (openinfo->flags & R_FIXEDLEN) {
			SET(t, R_FIXLEN);
			t->bt_reclen = openinfo->reclen;
			if (t->bt_reclen == 0)
				goto einval;
		}
		t->bt_bval = openinfo->bval;
	} else
		t->bt_bval = '\n';

	SET(t, R_RECNO);
	if (fname == NULL)
		SET(t, R_EOF | R_INMEM);
	else
#ifndef WIN32
		t->bt_rfd = rfd;
#else
		t->bt_Hrfd = Hrfd;
#endif
	t->bt_rcursor = 0;


        printf("rec_open: checking for pipe \n");
	/*
	 * In 4.4BSD stat(2) returns true for ISSOCK on pipes.  Until
	 * then, this is fairly close.  Pipes are read-only.
	 */
	if (fname != NULL) {
#ifndef WIN32
		if (lseek(rfd, (off_t)0, SEEK_CUR) == -1 && errno == ESPIPE) {
			switch (flags & O_ACCMODE) {
			case O_RDONLY:
				SET(t, R_RDONLY);
				break;
			default:
				goto einval;
			}
#else
                printf("rec_open: checking if %s is a pipe\n",fname);

                if (GetNamedPipeInfo(Hrfd,NULL,NULL,NULL,NULL)==TRUE) {
                    /* we must have a named pipe */
                    /* do nothing, since we don't care */
                    {  
                        printf("rec_open: was a pipe \n");
                    }
#endif

#ifndef WIN32
slow:			if ((t->bt_rfp = fdopen(rfd,"r")) == NULL)
				goto err;
			SET(t, R_CLOSEFP);
#else
slow:                   
                        printf("rec_open: opening rfp \n");
                        if ((t->bt_rfp = fopen(fname,"rb")) == NULL)
				goto err;
			SET(t, R_CLOSEFP);
#endif

			t->bt_irec =
			    ISSET(t, R_FIXLEN) ? __rec_fpipe : __rec_vpipe;
		} else {
			switch (flags & O_ACCMODE) {
			case O_RDONLY:
				SET(t, R_RDONLY);
				break;
			case O_RDWR:
				break;
			default:
				goto einval;
			}

#ifndef WIN32
			if (fstat(rfd, &sb))
				goto err;
#else
                        printf("rec_open:GetFileInformationByHandle\n");
                        GetFileInformationByHandle(Hrfd,
                                &BHFI);
#endif
			/*
			 * Kluge -- we'd like to test to see if the file is too
			 * big to mmap.  Since, we don't know what size or type
			 * off_t's or size_t's are, what the largest unsigned
			 * integral type is, or what random insanity the local
			 * C compiler will perpetrate, doing the comparison in
			 * a portable way is flatly impossible.  Hope that mmap
			 * fails if the file is too large.
			 */
#ifndef WIN32
			if (sb.st_size == 0)
#else
                        if ((BHFI.nFileSizeHigh | BHFI.nFileSizeLow) == 0)
#endif
				SET(t, R_EOF);
			else {
#ifndef WIN32
				t->bt_msize = sb.st_size;
#else
                        	t->bt_msize = BHFI.nFileSizeLow;
                                 printf("rec_open: file size %d \n",BHFI.nFileSizeLow);
#endif

#ifndef WIN32
				if ((t->bt_smap = mmap(NULL, t->bt_msize,
				    PROT_READ, MAP_PRIVATE, rfd,
				    (off_t)0)) == (caddr_t)-1)
					goto slow;
#else
                                /* Do Windows NT file mapping */
                                if ((Hfilemap = CreateFileMapping(Hrfd,NULL,
                                 PAGE_READWRITE,
                                    0,0,NULL)) != INVALID_HANDLE_VALUE) {
                                    printf("Couldn't MAP file\n");
                                    printf("Error code %d\n",
                                        GetLastError());
                                    goto err;
                                    }        

				/* map all of the file starting
                                at location zero */    
                                
                               if ((t->bt_smap =
                                    MapViewOfFile(Hfilemap,
                                    FILE_MAP_ALL_ACCESS,
                                       0,0,0)) == NULL) {
                                    printf("Couldn't Mapviewoffile\n");
                                    goto err;
                                }
#endif
				t->bt_cmap = t->bt_smap;
				t->bt_emap = t->bt_smap + BHFI.nFileSizeLow;
				t->bt_irec = ISSET(t, R_FIXLEN) ?
				    __rec_fmap : __rec_vmap;
				SET(t, R_MEMMAPPED);
			}

		}
	}

	/* Use the recno routines. */
	dbp->close = __rec_close;
	dbp->del = __rec_delete;
	dbp->fd = __rec_fd;
	dbp->get = __rec_get;
	dbp->put = __rec_put;
	dbp->seq = __rec_seq;
	dbp->sync = __rec_sync;

        printf("In rec_open - about to do mpool_get\n");

	/* If the root page was created, reset the flags. */
	if ((h = mpool_get(t->bt_mp, P_ROOT, 0)) == NULL)
		goto err;

        printf("In rec_open - about to do mpool_put\n");

	if ((h->flags & P_TYPE) == P_BLEAF) {
		h->flags = h->flags & ~P_TYPE | P_RLEAF;
		mpool_put(t->bt_mp, h, MPOOL_DIRTY);
	} else
		mpool_put(t->bt_mp, h, 0);

	if (openinfo && openinfo->flags & R_SNAPSHOT &&
	    !ISSET(t, R_EOF | R_INMEM) &&
	    t->bt_irec(t, MAX_REC_NUMBER) == RET_ERROR)
                goto err;
	return (dbp);

einval:	errno = EINVAL;
err:	sverrno = errno;
	if (dbp != NULL)
		(void)__bt_close(dbp);

	if (fname != NULL)
#ifndef WIN32
		(void)close(rfd);
#else
    		(void)CloseHandle(Hrfd);
#endif
	errno = sverrno;
	return (NULL);
}

#ifndef WIN32
int
#else
HANDLE
#endif
__rec_fd(dbp)
	const DB *dbp;
{
	BTREE *t;

	t = dbp->internal;

	/* Toss any page pinned across calls. */
	if (t->bt_pinned != NULL) {
		mpool_put(t->bt_mp, t->bt_pinned, 0);
		t->bt_pinned = NULL;
	}

	/* In-memory database can't have a file descriptor. */
	if (ISSET(t, R_INMEM)) {
		errno = ENOENT;
#ifndef WIN32
		return (-1);
#else
                return (INVALID_HANDLE_VALUE);
#endif
	}
#ifndef WIN32
	return (t->bt_rfd);
#else
        return (t->bt_Hrfd);
#endif
}

